@@ -23,6 +23,9 @@ hv.Store.renderers
 
 
 def get_coastline(ds, lon_name, lat_name):
+    """
+    A function to work out an appropriate coastline resolution
+    """
     lon_max = ds[lon_name].values.max()
     lon_min = ds[lon_name].values.min()
 
@@ -39,6 +42,9 @@ def get_coastline(ds, lon_name, lat_name):
 
 
 def change_coords(dx):
+    """
+    Some model output will have repeated zeros as lon/lat. This will mess up quadmesh. Replace lon/lat with indices
+    """
 
     ds = dx.copy()
     df_dims = get_dims(ds)
@@ -82,6 +88,10 @@ def change_coords(dx):
 
 
 def is_curvilinear(ds):
+    """
+    A function originally in nctoolkit, which used CDO to figure out if a grid was curvilinear. This should
+    be rename to something like is_quadmesh
+    """
 
     ds_dims = get_dims(ds)
     if ds_dims.longitude.values[0] is None or ds_dims.latitude.values[0] is None:
@@ -94,6 +104,9 @@ def is_curvilinear(ds):
 
 
 def ctrc():
+    """
+    A basic control to send a message to the terminal
+    """
     time.sleep(1)
     print("Press Ctrl+C to stop plotting server")
 
@@ -139,26 +152,59 @@ def view(x, vars=None):
     else:
         ds = x
 
+    coord_list = list(ds.coords)
+
+    for cc in coord_list:
+        if len(ds[cc].values.ravel()) <= 1:
+            if cc in list(ds.dims):
+                ds = ds.squeeze(cc, drop=True)
+
+    coord_list = list(ds.dims)
+
+    for cc in coord_list:
+        if len(ds[cc].values.ravel()) <= 1:
+            if cc in list(ds.dims):
+                ds = ds.squeeze(cc, drop=True)
+
     quadmesh = False
 
+    if len(list(ds.dims)) + len(list(ds.coords)) == 0:
+        raise ValueError("There are no dimensions or coordinates in the dataset!")
+
     df_dims = get_dims(ds)
     # figure out number of points
     lon_name = df_dims.longitude[0]
     lat_name = df_dims.latitude[0]
 
+    if lon_name is not None:
+        if "long_name" in ds[lon_name].attrs:
+            if "rotate" in ds[lon_name].long_name:
+                coastline = False
+
+            if "pole" in ds[lon_name].long_name:
+                coastline = False
+
+    if len([x for x in ds.coords if "lon" in x]) > len(
+        [x for x in ds.dims if "lon" in x]
+    ):
+        coastline = False
+
     if quadmesh is False and lat_name is not None:
         lats = ds[lat_name].values
+        if len(lats) > 1:
 
-        max_step = np.max([np.abs(x) for x in (lats[0:-2] - lats[1:-1])])
-        min_step = np.min([np.abs(x) for x in (lats[0:-2] - lats[1:-1])])
-        if min_step == 0:
-            quadmesh = True
-        else:
-            if max_step / min_step > 1.001:
+            max_step = np.max([np.abs(x) for x in (lats[0:-2] - lats[1:-1])])
+            min_step = np.min([np.abs(x) for x in (lats[0:-2] - lats[1:-1])])
+            if min_step == 0:
                 quadmesh = True
+            else:
+                if max_step / min_step > 1.001:
+                    quadmesh = True
 
     switch_coords = False
 
+    rasterize = True
+
     if quadmesh:
         switch_coords = True
         if len(np.shape(ds[lon_name].values)) == 1:
@@ -167,17 +213,14 @@ def view(x, vars=None):
         if switch_coords:
             if lon_name != check_lon(lon_name, ds):
                 lon_name = check_lon(lon_name, ds)
+                rasterize = False
                 switch_coords = False
+                quadmesh = False
                 coastline = False
             if lat_name != check_lat(lat_name, ds):
                 lat_name = check_lat(lat_name, ds)
-
-    if "long_name" in ds[lon_name]:
-        if "rotate" in ds[lon_name].long_name:
-            coastline = False
-
-        if "pole" in ds[lon_name].long_name:
-            coastline = False
+                rasterize = False
+                quadmesh = False
 
     if switch_coords:
         orig_coords = list(ds.coords)
@@ -190,6 +233,11 @@ def view(x, vars=None):
         lat_name = df_dims.latitude[0]
         if orig_coords != new_coords:
             coastline = False
+            rasterize = False
+
+    if lon_name is not None:
+        if len(ds[lon_name].dims) > 1:
+            quadmesh = True
 
     n_points = 1
     if lon_name is not None:
@@ -269,6 +317,26 @@ def view(x, vars=None):
     else:
         if vars not in list(ds.variables):
             raise ValueError(f"{vars} is not a valid variable")
+
+    if (lon_name is not None) and (lat_name is not None) and type(vars) is list:
+        new_vars = []
+        for x in vars:
+
+            dims_required = []
+            for i in list(ds.coords):
+                dims_required += ds.coords[i].dims
+            dims_required = set(dims_required)
+
+            if set(ds[x].dims) == dims_required:
+                new_vars.append(x)
+        vars = new_vars
+
+    if len(vars) == 1:
+        vars = vars[0]
+
+    if (lon_name is None) or (lat_name is None):
+        rasterize = False
+
     # Case when all you can plot is a time series
 
     # heat map 1
@@ -288,9 +356,19 @@ def view(x, vars=None):
         {"coord": coord_list, "length": [len(ds.coords[x].values) for x in coord_list]}
     )
 
+    # work out if it should be a spatial map
+
+    spatial_map = False
+
     # line plot
 
-    if len([x for x in coord_df.length if x > 1]) == 1:
+    if lon_name is not None and lat_name is not None:
+
+        if lon_name in list(ds.coords) and lat_name in list(ds.coords):
+            if (len(ds[lon_name].values) > 1) and (len(ds[lat_name].values) > 1):
+                spatial_map = True
+
+    if len([x for x in coord_df.length if x > 1]) == 1 and spatial_map is False:
 
         df = ds.to_dataframe().reset_index()
         if type(vars) is str:
@@ -322,7 +400,7 @@ def view(x, vars=None):
         return None
 
     # heat map where 2 coords have more than 1 value, not a spatial map
-    if len([x for x in coord_df.length if x > 1]) == 2:
+    if len([x for x in coord_df.length if x > 1]) == 2 and spatial_map is False:
 
         df = ds.to_dataframe().reset_index()
         x_var = coord_df.query("length > 1").reset_index().coord[0]
@@ -345,8 +423,32 @@ def view(x, vars=None):
         if case1 <= 1:
 
             if type(vars) is list:
+                # figure out if it needs to be quadmesh
+                if x_var == time_name:
+                    quadmesh = True
+                x_vals = ds[x_var].values
 
-                intplot = ds.hvplot.image(x_var, y_var, vars, cmap="viridis")
+                if len(x_vals) > 2:
+
+                    if (
+                        np.nanmax(x_vals[0:-2] - x_vals[1:-1]).astype("float")
+                        - np.nanmin(x_vals[0:-2] - x_vals[1:-1]).astype("float")
+                        > 0
+                    ):
+                        quadmesh = True
+                y_vals = ds[y_var].values
+                if len(y_vals) > 2:
+                    if (
+                        np.nanmax(y_vals[0:-2] - y_vals[1:-1]).astype("float")
+                        - np.nanmin(y_vals[0:-2] - y_vals[1:-1]).astype("float")
+                        > 0
+                    ):
+                        quadmesh = True
+
+                if quadmesh:
+                    intplot = ds.hvplot.quadmesh(x_var, y_var, vars, cmap="viridis")
+                else:
+                    intplot = ds.hvplot.image(x_var, y_var, vars, cmap="viridis")
 
             else:
 
@@ -354,18 +456,52 @@ def view(x, vars=None):
                 self_min = ds.rename({vars: "x"}).x.min()
                 v_max = float(max(self_max.values, -self_min.values))
 
+                # figure out if it needs to be quadmesh
+                if x_var == time_name:
+                    quadmesh = True
+                x_vals = ds[x_var].values
+
+                if len(x_vals) > 2:
+                    if (
+                        np.nanmax(x_vals[0:-2] - x_vals[1:-1]).astype("float")
+                        - np.nanmin(x_vals[0:-2] - x_vals[1:-1]).astype("float")
+                        > 0
+                    ):
+                        quadmesh = True
+                y_vals = ds[y_var].values
+
+                if len(y_vals) > 2:
+                    if (
+                        np.nanmax(y_vals[0:-2] - y_vals[1:-1]).astype("float")
+                        - np.nanmin(y_vals[0:-2] - y_vals[1:-1]).astype("float")
+                        > 0
+                    ):
+                        quadmesh = True
+
                 if self_max > 0 and self_min < 0:
-                    intplot = ds.hvplot.image(
-                        x_var,
-                        y_var,
-                        vars,
-                        cmap="RdBu_r",
-                        rasterize=True,
-                        responsive=(in_notebook() is False),
-                    ).redim.range(**{vars: (-v_max, v_max)})
+                    if quadmesh:
+                        intplot = ds.hvplot.quadmesh(
+                            x_var,
+                            y_var,
+                            vars,
+                            cmap="RdBu_r",
+                            responsive=(in_notebook() is False),
+                        ).redim.range(**{vars: (-v_max, v_max)})
+                    else:
+                        intplot = ds.hvplot.image(
+                            x_var,
+                            y_var,
+                            vars,
+                            cmap="RdBu_r",
+                            rasterize=False,
+                            responsive=(in_notebook() is False),
+                        ).redim.range(**{vars: (-v_max, v_max)})
 
                 else:
-                    intplot = ds.hvplot.image(x_var, y_var, vars, cmap="viridis")
+                    if quadmesh:
+                        intplot = ds.hvplot.quadmesh(x_var, y_var, vars, cmap="viridis")
+                    else:
+                        intplot = ds.hvplot.image(x_var, y_var, vars, cmap="viridis")
 
             if in_notebook():
                 return intplot
@@ -405,9 +541,6 @@ def view(x, vars=None):
                 time_in = True
                 possible += 1
 
-        # if possible > 1:
-        #    time_in = False
-
         if time_name in coord_list and time_in and non_map:
 
             if coord_df.query("coord == @time_name").length.values > 1:
@@ -446,14 +579,37 @@ def view(x, vars=None):
 
                 if case1 <= 1:
 
-                    intplot = df.drop_duplicates().hvplot.heatmap(
-                        x=x_var,
-                        y=y_var,
-                        C="value",
-                        groupby=["variable", time_name],
-                        cmap="viridis",
-                        colorbar=True,
-                    )
+                    # this could also do with a blue to red option
+
+                    # figure out if it should be quadmesh
+                    quadmesh = False
+                    x_vals = ds[x_var].values
+                    if len(x_vals) > 2:
+                        if (
+                            np.nanmax(x_vals[0:-2] - x_vals[1:-1]).astype("float")
+                            - np.nanmin(x_vals[0:-2] - x_vals[1:-1]).astype("float")
+                            > 0
+                        ):
+                            quadmesh = True
+
+                    y_vals = ds[y_var].values
+                    if len(y_vals) > 2:
+                        if (
+                            np.nanmax(y_vals[0:-2] - y_vals[1:-1]).astype("float")
+                            - np.nanmin(y_vals[0:-2] - y_vals[1:-1]).astype("float")
+                            > 0
+                        ):
+                            quadmesh = True
+
+                    if quadmesh:
+                        intplot = ds.hvplot.quadmesh(
+                            x_var, y_var, vars, cmap="viridis", rasterize=True
+                        )
+                    else:
+                        intplot = ds.hvplot.image(
+                            x_var, y_var, vars, cmap="viridis", rasterize=False
+                        )
+
                     if in_notebook():
                         return intplot
 
@@ -520,9 +676,6 @@ def view(x, vars=None):
             else:
                 projection = None
 
-            print(lon_name)
-            print(lat_name)
-            print(vars)
             intplot = ds.hvplot.quadmesh(
                 lon_name,
                 lat_name,
@@ -531,7 +684,7 @@ def view(x, vars=None):
                 cmap="viridis",
                 coastline=coastline,
                 projection=projection,
-                rasterize=True,
+                rasterize=rasterize,
                 responsive=in_notebook() is False,
             )
         else:
@@ -541,6 +694,7 @@ def view(x, vars=None):
                 projection = ccrs.PlateCarree()
             else:
                 projection = None
+
             intplot = ds.hvplot.image(
                 lon_name,
                 lat_name,
@@ -548,6 +702,7 @@ def view(x, vars=None):
                 dynamic=True,
                 cmap="viridis",
                 coastline=coastline,
+                rasterize=False,
                 projection=projection,
                 responsive=in_notebook() is False,
             )
@@ -568,8 +723,8 @@ def view(x, vars=None):
             warnings.warn(message="Warning: Only the first variable is mapped")
             vars = vars[0]
 
-        self_max = ds.rename({vars: "x"}).x.max()
-        self_min = ds.rename({vars: "x"}).x.min()
+        self_max = ds.rename({vars: "x_dim12345"}).x_dim12345.max()
+        self_min = ds.rename({vars: "x_dim12345"}).x_dim12345.min()
         v_max = max(self_max.values, -self_min.values)
 
         if (self_max.values > 0) and (self_min.values < 0):
@@ -588,7 +743,7 @@ def view(x, vars=None):
                     cmap="RdBu_r",
                     coastline=coastline,
                     projection=projection,
-                    rasterize=True,
+                    rasterize=rasterize,
                     responsive=(in_notebook() is False),
                 ).redim.range(**{vars: (-v_max, v_max)})
                 # intplot = pn.Row(pn.WidgetBox(w1), intplot)
@@ -606,6 +761,7 @@ def view(x, vars=None):
                     dynamic=True,
                     coastline=coastline,
                     projection=projection,
+                    rasterize=False,
                     cmap="RdBu_r",
                     responsive=(in_notebook() is False),
                 ).redim.range(**{vars: (-v_max, v_max)})
@@ -635,7 +791,7 @@ def view(x, vars=None):
                     cmap="viridis",
                     coastline=coastline,
                     projection=projection,
-                    rasterize=True,
+                    rasterize=rasterize,
                     responsive=(in_notebook() is False),
                 ).redim.range(**{vars: (self_min.values, v_max)})
             else:
@@ -653,6 +809,7 @@ def view(x, vars=None):
                     dynamic=True,
                     cmap="viridis",
                     coastline=coastline,
+                    rasterize=False,
                     projection=projection,
                     responsive=(in_notebook() is False),
                 ).redim.range(**{vars: (self_min.values, v_max)})
@@ -668,6 +825,8 @@ def view(x, vars=None):
 
             return None
 
+    # it's possible that the data has not spatial or temporal coordinates. Do something about it....
+
     # Throw an error if case has not plotting method available yet
     # right now this only seems to be when you have lon/lat/time/levels and multiple variables
     # maybe needs an appropriate
